from io import BytesIO

import requests
from lxml import etree

from ckan.plugins import toolkit

parser = etree.XMLParser(recover=True)

standard_fields = ['datasetID', 'basisOfRecord', 'dynamicProperties']
valid_types = ['StillImage', 'MovingImage', 'Sound', 'PhysicalObject', 'Event', 'Text']


def load_schema(url, base_url=None):
    # lxml does not like https urls, so use requests to get the bytes
    r = requests.get(url)
    content = BytesIO(r.content)
    return etree.parse(content, base_url=base_url).getroot()


def json_to_xml(tag, content):
    if isinstance(content, tuple):
        attrs = content[1]
        content = content[0]
    else:
        attrs = {}

    if isinstance(content, dict):
        node = etree.Element(tag, **attrs)
        for k, v in content.items():
            for n in json_to_xml(k, v):
                node.append(n)
        return [node]
    elif isinstance(content, list):
        nodes = [item for e in content for item in json_to_xml(tag, e)]
        return nodes
    else:
        node = etree.Element(tag, **attrs)
        node.text = str(content)
        return [node]


def get_setting(*config_names, default=None):
    setting = None
    for c in config_names:
        setting = toolkit.config.get(c)
        if setting is not None:
            break
    return setting or default


class NSMap(dict):
    def ns(self, key, tag):
        return f'{{{self[key]}}}{tag}'



"""
Add navigational slugs.

Revision ID: 526b12c69d55
Revises: 19a61e5b669f
Create Date: 2023-06-07 16:25:59.090795
"""
from datetime import datetime as dt
from uuid import uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base

# revision identifiers, used by Alembic.
revision = '526b12c69d55'
down_revision = '19a61e5b669f'
branch_labels = None
depends_on = None

Base = declarative_base()


def make_uuid():
    return str(uuid4())


class NavigationalSlug(Base):
    __tablename__ = 'versioned_datastore_navigational_slugs'

    id = sa.Column(sa.UnicodeText, primary_key=True, default=make_uuid)
    query_hash = sa.Column(sa.UnicodeText, nullable=False, index=True, unique=True)
    query = sa.Column(JSONB, nullable=False)
    resource_ids_and_versions = sa.Column(JSONB, nullable=False, default=dict)
    created = sa.Column(sa.DateTime, nullable=False, default=dt.utcnow)


def upgrade():
    bind = op.get_bind()
    NavigationalSlug.__table__.create(bind)


def downgrade():
    bind = op.get_bind()
    NavigationalSlug.__table__.drop(bind)



import copy
import logging

from ckan.plugins import PluginImplementations
from splitgill.indexing.feeders import SimpleIndexFeeder
from splitgill.indexing.indexers import Indexer
from splitgill.indexing.indexes import Index
from splitgill.indexing.utils import get_versions_and_data, DOC_TYPE

from . import stats
from .. import common
from ...interfaces import IVersionedDatastore

log = logging.getLogger(__name__)


class DatastoreIndex(Index):
    """
    Represents an index in elasticsearch for a resource in CKAN for the splitgill
    indexing process.
    """

    def __init__(
        self, config, name, version, latitude_field=None, longitude_field=None
    ):
        '''
        :param config: the splitgill config object
        :param name: the resource id, this will be used as the index name
        :param version: the version being indexed up to
        :param latitude_field: optional - the name of a field containing latitudinal data
        :param longitude_field: optional - the name of a field containing longitudinal data
        '''
        super(DatastoreIndex, self).__init__(config, name, version)
        self.latitude_field = latitude_field
        self.longitude_field = longitude_field

    def add_geo_data(self, index_doc):
        """
        Adds a geo point to the meta part of the index document. This is done in place.
        If the latitude and longitude fields have not been specified by the user, are
        not present in the data then or have invalid values then nothing happens.

        :param index_doc: the dict to be indexed
        """
        if self.latitude_field and self.longitude_field:
            # extract the latitude and longitude values
            latitude = index_doc['data'].get(self.latitude_field, None)
            longitude = index_doc['data'].get(self.longitude_field, None)
            if latitude is not None and longitude is not None:
                try:
                    # check that the values are valid
                    if -90 <= float(latitude) <= 90 and -180 <= float(longitude) <= 180:
                        # update the meta.geo key to hold the latitude longitude pair
                        index_doc['meta']['geo'] = f'{latitude},{longitude}'
                except ValueError:
                    pass

    def get_index_docs(self, mongo_doc):
        """
        Yields all the action and data dicts as a tuple for the given mongo doc. To make
        things consistent the id of the record is copied into the _id field when the
        record is indexed. This ensures that everything in the datastore index has this
        field.

        :param mongo_doc: the mongo doc to handle
        """
        # iterate over the mongo_docs versions and send them to elasticsearch
        for version, data, next_version in get_versions_and_data(mongo_doc):
            # if the data is empty, skip it, it's probably a deletion
            if not data:
                continue
            # copy the dict's data so that we can modify it safely (see get_versions_and_data doc)
            to_index = copy.deepcopy(data)
            # copy over the id of the record into the correctly named field and convert it to an
            # integer
            to_index['_id'] = int(mongo_doc['id'])

            for key, value in to_index.items():
                if isinstance(value, bool):
                    # boolean values when passed through to elasticsearch will be rejected as they
                    # currently aren't handled silently when converting to a number fails, therefore
                    # just convert it to a string here. Note that records with boolean values can
                    # only come from the API as if you upload a CSV we won't convert the value to
                    # a boolean ever
                    to_index[key] = str(value).lower()
                else:
                    # everything else can pass on through
                    to_index[key] = value

            # create the base index doc
            index_doc = self.create_index_document(to_index, version, next_version)

            # add geo data if there is any to add
            self.add_geo_data(index_doc)

            # allow other extensions implementing our interface to modify the index doc
            for plugin in PluginImplementations(IVersionedDatastore):
                index_doc = plugin.datastore_modify_index_doc(
                    self.unprefixed_name, index_doc
                )

            # yield it
            yield version, index_doc

    def get_index_create_body(self):
        """
        Returns a dict that should be used to define the index in elasticsearch. This
        should be built off of the default splitgill one unless you super know what
        you're doing.

        :return: a dict
        """
        body = super(DatastoreIndex, self).get_index_create_body()
        body.setdefault('mappings', {}).setdefault(DOC_TYPE, {}).setdefault(
            'properties', {}
        ).update(
            {
                # the id field should be an integer
                'data._id': {'type': 'long'},
            }
        )
        return body


class ResourceIndexRequest(object):
    """
    Class representing a request to index data for a resource.

    We use a class like this to avoid having a long list of arguments passed through to
    queued functions.
    """

    def __init__(self, resource, lower_version, upper_version):
        '''
        :param resource: the dict for the resource we're going to index
        :param lower_version: the lower version to index (exclusive)
        :param upper_version: the upper version to index (inclusive)
        '''
        self.resource = resource
        self.lower_version = lower_version
        self.upper_version = upper_version
        self.resource_id = self.resource['id']

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f'Index on {self.resource_id}, lower version: {self.lower_version}, '
            f'upper version: {self.upper_version}'
        )


def index_resource(request):
    """
    Indexes a resource's data from MongoDB into Elasticsearch.

    :param request: the request
    :return: True if the index took place successfully, False if not
    """
    resource_id = request.resource['id']

    log.info(
        f'Starting index of {resource_id}: {request.lower_version} > versions <= '
        f'{request.upper_version}'
    )

    # create an index feeder, this gets the records out of MongoDB and presents them for indexing
    feeder = SimpleIndexFeeder(
        common.CONFIG, resource_id, request.lower_version, request.upper_version
    )
    # create the index object which will process the documents from the feeder
    index = DatastoreIndex(
        common.CONFIG,
        resource_id,
        request.upper_version,
        latitude_field=request.resource.get('_latitude_field', None),
        longitude_field=request.resource.get('_longitude_field', None),
    )
    # create an indexer object to do the indexing
    indexer = Indexer(request.upper_version, common.CONFIG, [(feeder, index)])

    # create a stats entry so that progress can be tracked
    stats_id = stats.start_operation(
        resource_id, stats.INDEX, request.upper_version, indexer.start
    )
    # setup monitoring on the indexer so that we can update the database with stats about the
    # index operation as it progresses
    stats.monitor_indexing(stats_id, indexer)

    try:
        # actually do the index, getting back some stats
        indexing_stats = indexer.index()
    except Exception as e:
        # if there's a problem, mark it in the stats
        stats.mark_error(stats_id, e)
        log.exception(f'An error occurred during indexing of {resource_id}')
        return False

    # otherwise, we're all good, let the plugins do stuff if they want
    for plugin in PluginImplementations(IVersionedDatastore):
        try:
            plugin.datastore_after_indexing(request, indexing_stats, stats_id)
        except Exception as e:
            log.error(
                f'Error during after indexing hook handling in plugin {plugin}', e
            )
    # done, return True to indicate all went successfully
    return True



from datetime import datetime

from ckan.plugins import toolkit, PluginImplementations
from splitgill.indexing.utils import DOC_TYPE
from splitgill.search import create_version_query
from splitgill.utils import to_timestamp
from elasticsearch import RequestError
from elasticsearch_dsl import A, Search

from .meta import help, schema
from ckantools.decorators import action
from ...interfaces import IVersionedDatastore
from ...lib.basic_query.search import create_search
from ...lib.basic_query.utils import run_search, get_fields, format_facets
from ...lib.datastore_utils import prefix_field, prefix_resource, get_last_after
from ...lib.query.query_log import log_query


# these are the keys we're interested in logging from the data dict
_query_log_keys = ('q', 'filters')


@action(schema.datastore_search(), help.datastore_search, get=True)
def datastore_search(context, data_dict, original_data_dict):
    """
    Searches the datastore using a query schema similar to the standard CKAN datastore
    query schema, but with versioning.

    :param context: the context dict from the action call
    :param data_dict: the data_dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :return: a dict including the search results amongst other things
    """
    original_data_dict, data_dict, version, search = create_search(
        context, data_dict, original_data_dict
    )
    resource_id = data_dict['resource_id']
    index_name = prefix_resource(resource_id)

    # if the version is None, default it to the current timestamp
    if version is None:
        version = to_timestamp(datetime.now())

    # add the version filter to the query
    search = search.filter(create_version_query(version))

    # if the run query option is false (default to true if not present) then just return the query
    # we would have run against elasticsearch instead of actually running it. This is useful for
    # running the query outside of ckan, for example on a tile server.
    if not data_dict.get('run_query', True):
        return {
            'indexes': [index_name],
            'search': search.to_dict(),
        }
    else:
        result = run_search(search, [index_name])

        # allow other extensions implementing our interface to modify the result
        for plugin in PluginImplementations(IVersionedDatastore):
            result = plugin.datastore_modify_result(
                context, original_data_dict, data_dict, result
            )

        # add the actual result object to the context in case the caller is an extension and they
        # have used one of the interface hooks to alter the search object and include, for example,
        # an aggregation
        context['versioned_datastore_query_result'] = result

        # get the fields
        mapping, fields = get_fields(resource_id, version)
        # allow other extensions implementing our interface to modify the field definitions
        for plugin in PluginImplementations(IVersionedDatastore):
            fields = plugin.datastore_modify_fields(resource_id, mapping, fields)

        query_for_logging = {}
        for key in _query_log_keys:
            if data_dict.get(key, None):
                query_for_logging[key] = data_dict[key]
        log_query(query_for_logging, 'basicsearch')

        # return a dictionary containing the results and other details
        return {
            'total': result.hits.total,
            'records': [hit.data.to_dict() for hit in result],
            'facets': format_facets(result.aggs.to_dict()),
            'fields': fields,
            'raw_fields': mapping['mappings'][DOC_TYPE]['properties']['data'][
                'properties'
            ],
            'after': get_last_after(result.hits),
            '_backend': 'versioned-datastore',
        }


@action(
    schema.datastore_autocomplete(),
    help.datastore_autocomplete,
    toolkit.side_effect_free,
)
def datastore_autocomplete(context, data_dict, original_data_dict):
    """
    Runs a search to find autocomplete values based on the provided prefix.

    :param context: the context dict from the action call
    :param data_dict: the data_dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :return: a dict containing a list of results and an after value for the next page of results
    """
    # extract the fields specifically needed for setting up the autocomplete query
    field = data_dict.pop('field')
    term = data_dict.pop('term')
    after = data_dict.pop('after', None)
    # default to a size of 20 results
    size = data_dict.pop('limit', 20)
    # ensure the search doesn't respond with any hits cause we don't need them
    data_dict['limit'] = 0
    # remove the offset if one was passed as we don't need it
    data_dict.pop('offset', None)

    # now build the search object against the normal search code
    original_data_dict, data_dict, version, search = create_search(
        context, data_dict, original_data_dict
    )
    # get the index we're going to search against
    index_name = prefix_resource(data_dict['resource_id'])

    # add the autocompletion query part which takes the form of a prefix search
    search = search.filter('prefix', **{prefix_field(field): term})
    # modify the search so that it has the aggregation required to get the autocompletion results
    search.aggs.bucket(
        'field_values',
        'composite',
        size=size,
        sources={field: A('terms', field=prefix_field(field), order='asc')},
    )
    # if there's an after included, add it into the aggregation
    if after:
        search.aggs['field_values'].after = {field: after}

    # run the search (this adds the version to the query too)
    result = run_search(search, [index_name], version)

    # get the results we're interested in
    agg_result = result.aggs.to_dict()['field_values']
    # return a dict of results, but only include the after details if there are any to include
    return_dict = {
        'values': [bucket['key'][field] for bucket in agg_result['buckets']],
    }
    if 'after_key' in agg_result:
        return_dict['after'] = agg_result['after_key'][field]
    return return_dict


@action(
    schema.datastore_search(), help.datastore_query_extent, toolkit.side_effect_free
)
def datastore_query_extent(context, data_dict, original_data_dict):
    """
    Given the parameters for a datastore_query, finds the geographic extent of the
    query's results.

    :param context: the context dict from the action call
    :param data_dict: the data_dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :return: a dict containing the total number of matches for the query, the total number of
             matches with geo data and the bounds of the query
    """
    # ensure the search doesn't respond with any hits cause we don't need them and override two
    # unused params
    data_dict['limit'] = 0
    data_dict.pop('offset', None)
    data_dict.pop('after', None)

    # now build the search object against the normal search code
    original_data_dict, data_dict, version, search = create_search(
        context, data_dict, original_data_dict
    )
    # if we don't have a version, set to now
    if version is None:
        version = to_timestamp(datetime.now())

    # get the index we're going to search against
    index_name = prefix_resource(data_dict['resource_id'])

    # add our bounds and geo count aggregations
    search.aggs.bucket('bounds', 'geo_bounds', field='meta.geo', wrap_longitude=False)
    search.aggs.bucket('geo_count', 'value_count', field='meta.geo')

    # add version filter and run the search
    result = run_search(search, [index_name], version)
    agg_result = result.aggs.to_dict()

    # create a dict of results for return
    to_return = {
        'total_count': result.hits.total,
        'geom_count': agg_result['geo_count']['value'],
    }

    # extract and add the bounds info from the aggregations if there is any
    if agg_result['geo_count']['value'] > 0:
        top_left = agg_result['bounds']['bounds']['top_left']
        bottom_right = agg_result['bounds']['bounds']['bottom_right']
        to_return['bounds'] = [[p['lat'], p['lon']] for p in (top_left, bottom_right)]

    return to_return


@action(
    schema.datastore_search_raw(), help.datastore_search_raw, toolkit.side_effect_free
)
def datastore_search_raw(
    resource_id,
    context,
    data_dict,
    original_data_dict,
    search=None,
    version=None,
    raw_result=False,
    include_version=True,
):
    """
    Searches the datastore using a raw elasticsearch query.

    :param resource_id: the id of the resource to search
    :param context: the context dict from the action call
    :param data_dict: the data_dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :param search: the elasticsearch query to run
    :param version: the version of the data to query against
    :param raw_result: whether to return the result as a raw elasticsearch result, or format it in
                       the same way as a normal datastore_search call would
    :param include_version: whether to include the version in the search or not
    :return: a dict containing the results of the search
    """
    if search is None:
        search = {}
    if version is None:
        version = to_timestamp(datetime.now())
    index_name = prefix_resource(resource_id)
    search = Search.from_dict(search)

    try:
        # the user has asked for a raw result and that the version filter is not included
        if raw_result and not include_version:
            version = None

        # run the query passing the version which will either be the requested version, the current
        # timestamp or None if no version filter should be included in the search
        result = run_search(search, index_name, version)

        if raw_result:
            return result.to_dict()

        # allow other extensions implementing our interface to modify the result object
        for plugin in PluginImplementations(IVersionedDatastore):
            result = plugin.datastore_modify_result(
                context, original_data_dict, data_dict, result
            )

        # add the actual result object to the context in case the caller is an extension and
        # they have used one of the interface hooks to alter the search object and include, for
        # example, an aggregation
        context['versioned_datastore_query_result'] = result

        # get the fields
        mapping, fields = get_fields(resource_id, version)
        # allow other extensions implementing our interface to modify the field definitions
        for plugin in PluginImplementations(IVersionedDatastore):
            fields = plugin.datastore_modify_fields(resource_id, mapping, fields)

        # return a dictionary containing the results and other details
        return {
            'total': result.hits.total,
            'records': [hit.data.to_dict() for hit in result],
            'facets': format_facets(result.aggs.to_dict()),
            'fields': fields,
            'raw_fields': mapping['mappings'][DOC_TYPE]['properties']['data'][
                'properties'
            ],
            'after': get_last_after(result.hits),
            '_backend': 'versioned-datastore',
        }
    except RequestError as e:
        raise toolkit.ValidationError(str(e))



from ckan.plugins import toolkit
from datetime import datetime
from splitgill.utils import to_timestamp

from .meta import help, schema
from ckantools.decorators import action
from ...lib import common
from ...lib.datastore_utils import (
    is_resource_read_only,
    is_ingestible,
    update_privacy,
    ReadOnlyResourceException,
    InvalidVersionException,
    is_datastore_resource,
)
from ...lib.importing import stats
from ...lib.importing.indexing import DatastoreIndex
from ...lib.importing.queuing import queue_index, queue_import, queue_deletion
from ...lib.importing.utils import check_version_is_valid


@action(schema.datastore_create(), help.datastore_create)
def datastore_create(resource_id, context):
    """
    Creates an index in elasticsearch for the given resource unless it's a read only
    index or it looks uningestible.

    :param resource_id: the resource's id
    :param context: the context dict from the action call
    :return: True if the index was created or already existed, False if not
    """
    if is_resource_read_only(resource_id):
        return False

    # lookup the resource dict
    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    # only create the index if the resource is ingestable
    if is_ingestible(resource):
        # note that the version parameter doesn't matter when creating the index so we can safely
        # pass None
        common.SEARCH_HELPER.ensure_index_exists(
            DatastoreIndex(common.CONFIG, resource_id, None)
        )
        # make sure the privacy is correctly setup
        update_privacy(resource_id)
        return True
    return False


@action(schema.datastore_upsert(), help.datastore_upsert)
def datastore_upsert(resource_id, replace, context, original_data_dict, version=None):
    """
    Main data ingestion function for the datastore. The URL on the given resource will
    be used to retrieve and then ingest data or, if provided, records will be ingested
    directly from the request. Data is ingested using the an rq background job and
    therefore this is an async action.

    :param resource_id: the resource to ingest the data into
    :param replace: whether to replace the data already in the resource or append to it
    :param context: the context dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :param version: the version of the new data, can be None (default) but if not must be newer
                    than the latest version of the resource
    :return: information about the background job that is handling the ingestion
    """
    # this comes through as junk if it's not removed before validating. This happens because the
    # data dict is flattened during validation, but why this happens is unclear.
    records = original_data_dict.get('records', None)

    if is_resource_read_only(resource_id):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    if version is None:
        version = to_timestamp(datetime.now())

    # check that the version is valid
    if not check_version_is_valid(resource_id, version):
        raise InvalidVersionException(
            'The new version must be newer than current version'
        )

    # get the current user
    user = toolkit.get_action('user_show')(context, {'id': context['user']})

    # queue the resource import job
    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    job = queue_import(resource, version, replace, records, user['apikey'])

    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
    }


@action(schema.datastore_delete(), help.datastore_delete)
def datastore_delete(resource_id, context, version=None):
    """
    Deletes the resource from the datastore. In reality the resource data is maintained
    in its index but the latest version of all records is set to an empty record. This
    means that the old data is still accessible to ensure searches using versions before
    the deletion still work but searches at the latest version or later will return no
    records. The deletion work is done by an rq background job and therefore this is an
    async action.

    :param resource_id: the id of the resource to delete
    :param context: the context dict from the action call
    :param version: the to mark the deletion at
    :return: a dict containing info about the background job that is doing the delete
    """
    # if the requested deletion version is missing, default to now
    if version is None:
        version = to_timestamp(datetime.now())

    if is_resource_read_only(resource_id):
        raise toolkit.ValidationError('This resource has been marked as read only')

    # queue the job
    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    job = queue_deletion(resource, version)
    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
    }


@action(schema.datastore_reindex(), help.datastore_reindex)
def datastore_reindex(resource_id, context):
    """
    Reindexes that data in the given resource, this involves recreating all versions of
    each record in elasticsearch using the data in mongo. The reindexing work is done by
    an rq background job and therefore this is an async action.

    :param resource_id: the resource id to reindex
    :param context: the context dict from the action call
    :return: a dict containing info about the background job that is doing the reindexing
    """
    if is_resource_read_only(resource_id):
        raise toolkit.ValidationError('This resource has been marked as read only')

    last_ingested_version = stats.get_last_ingest(resource_id)
    if last_ingested_version is None:
        raise toolkit.ValidationError('There is no ingested data for this version')

    resource = toolkit.get_action('resource_show')(context, {'id': resource_id})
    job = queue_index(resource, None, last_ingested_version.version)

    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
    }


@action(schema.datastore_ensure_privacy(), help.datastore_ensure_privacy)
def datastore_ensure_privacy(context, resource_id=None):
    """
    Ensures that the privacy settings for the given resource, or if not resource is
    provided all resources, are correct. This means ensuring that the public alias for
    the resource's index exists or doesn't depending on the owning package's privacy
    setting.

    :param context: the context dict from the action call
    :param resource_id: the id of the resource update. Can be None (the default) which means all
                        resources are updated
    :return: a dict containing the total number of resources checked and the total modified
    """
    modified = 0
    total = 0
    if resource_id is not None:
        total += 1
        if is_datastore_resource(resource_id):
            if update_privacy(resource_id):
                modified += 1
    else:
        package_data_dict = {'limit': 50, 'offset': 0}
        while True:
            # iteratively retrieve all packages and ensure their resources
            packages = toolkit.get_action('current_package_list_with_resources')(
                context, package_data_dict
            )
            if not packages:
                # we've ensured all the packages that are available
                break
            else:
                # setup the next loop so that we get the next page of results
                package_data_dict['offset'] += len(packages)
                for package in packages:
                    for resource in package.get('resources', []):
                        if resource['datastore_active']:
                            total += 1
                            if update_privacy(resource['id'], package['private']):
                                modified += 1

    return {'modified': modified, 'total': total}



from dataclasses import dataclass


@dataclass
class SchemaUrl:
    url: str
    base: str
    fields: list = None


class TDWGUrls:
    base_url = 'https://dwc.tdwg.org/xml'
    xmlns = 'http://rs.tdwg.org/dwc/text'
    core = 'https://dwc.tdwg.org/xml/tdwg_dwcterms.xsd'
    base_types = 'https://dwc.tdwg.org/xml/tdwg_basetypes.xsd'
    core_classes = 'https://dwc.tdwg.org/xml/tdwg_dwc_classes.xsd'
    core_class_terms = 'https://dwc.tdwg.org/xml/tdwg_dwc_class_terms.xsd'
    simple = 'https://dwc.tdwg.org/xml/tdwg_dwc_simple.xsd'
    terms_csv = (
        'https://raw.githubusercontent.com/tdwg/dwc/master/vocabulary/term_versions.csv'
    )
    metadata = 'http://rs.tdwg.org/dwc/text/tdwg_dwc_text.xsd'


class GBIFUrls:
    base_url = 'https://rs.gbif.org'
    eml = 'http://rs.gbif.org/schema/eml-gbif-profile/1.1/eml.xsd'
    thesaurus = 'http://rs.gbif.org/vocabulary/gbif/dataset_type.xml'


class XMLUrls:
    xsi = 'http://www.w3.org/2001/XMLSchema-instance'
    xs = 'http://www.w3.org/2001/XMLSchema'
    xml = 'http://www.w3.org/XML/1998/namespace'
    dc = 'http://purl.org/dc/terms'
    eml = 'eml://ecoinformatics.org/eml-2.1.1'


core_extensions = {
    'gbif_occurrence': SchemaUrl(
        'https://rs.gbif.org/core/dwc_occurrence_2020-07-15.xml', GBIFUrls.base_url
    ),
    'gbif_taxon': SchemaUrl(
        'https://rs.gbif.org/core/dwc_taxon_2015-04-24.xml', GBIFUrls.base_url
    ),
    'gbif_event': SchemaUrl(
        'https://rs.gbif.org/core/dwc_event_2016_06_21.xml', GBIFUrls.base_url
    ),
}

extensions = {
    'gbif_multimedia': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/multimedia.xml',
        GBIFUrls.base_url,
        ['associatedMedia'],
    ),
    'gbif_vernacular': SchemaUrl(
        'http://rs.gbif.org/extension/gbif/1.0/vernacularname.xml',
        GBIFUrls.base_url,
        ['vernacularName'],
    ),
    'gbif_references': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/references.xml',
        GBIFUrls.base_url,
        ['references'],
    ),
    'gbif_description': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/description.xml',
        GBIFUrls.base_url,
        ['taxonDescription'],  # not a standard field
    ),
    'gbif_distribution': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/distribution.xml',
        GBIFUrls.base_url,
        ['locality'],  # not a standard Taxon field but is a valid DwC field
    ),
    'gbif_species_profile': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/speciesprofile.xml',
        GBIFUrls.base_url,
        ['speciesProfile'],  # not a standard field
    ),
    'gbif_types_and_specimen': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/typesandspecimen.xml',
        GBIFUrls.base_url,
        ['typeStatus'],
    ),
    'gbif_identifier': SchemaUrl(
        'https://rs.gbif.org/extension/gbif/1.0/identifier.xml',
        GBIFUrls.base_url,
        ['verbatimIdentification'],  # an Identification field
    ),
}



import io
import json
import os
from collections import OrderedDict

import jsonschema
import pytest
from ckanext.versioned_datastore.lib.query import schema as schema_lib
from mock import MagicMock, patch


class TestQuery(object):
    def test_load_core_schema(self):
        schema, validator = schema_lib.load_core_schema('v1.0.0')

        schema_path = os.path.join(schema_lib.schema_base_path, 'v1.0.0', 'v1.0.0.json')
        with io.open(schema_path, 'r', encoding='utf-8') as f:
            assert schema == json.load(f)
        # the validator classes are created on the fly so we'll just assert that the returned
        # validator object has a validate method on it as that's all we really care about
        assert hasattr(validator, 'validate')

    def test_schema_base_path_is_absolute(self):
        # load_core_schema needs this to be absolute
        assert os.path.isabs(schema_lib.schema_base_path)

    def test_get_latest_query_version(self):
        test_schemas = OrderedDict(
            [
                ('v1.0.0', MagicMock()),
                ('v1.0.1', MagicMock()),
                ('v2.10.1', MagicMock()),
            ]
        )
        with patch(
            'ckanext.versioned_datastore.lib.query.schema.schemas', test_schemas
        ):
            assert schema_lib.get_latest_query_version() == 'v1.0.0'

    def test_validate_query_valid(self):
        test_schemas = {'v1.0.0': MagicMock(validate=MagicMock(return_value=True))}
        with patch(
            'ckanext.versioned_datastore.lib.query.schema.schemas', test_schemas
        ):
            assert schema_lib.validate_query(MagicMock(), 'v1.0.0')

    def test_validate_query_invalid_query_version(self):
        with patch('ckanext.versioned_datastore.lib.query.schema.schemas', {}):
            with pytest.raises(schema_lib.InvalidQuerySchemaVersionError):
                schema_lib.validate_query(MagicMock(), 'v3.0.0')

    def test_validate_query_invalid_query(self):
        test_schemas = {
            'v1.0.0': MagicMock(
                validate=MagicMock(side_effect=jsonschema.ValidationError('no!'))
            )
        }
        with patch(
            'ckanext.versioned_datastore.lib.query.schema.schemas', test_schemas
        ):
            with pytest.raises(jsonschema.ValidationError):
                schema_lib.validate_query(MagicMock(), 'v1.0.0')

    def test_translate_query_invalid(self):
        with patch('ckanext.versioned_datastore.lib.query.schema.schemas', {}):
            with pytest.raises(schema_lib.InvalidQuerySchemaVersionError):
                schema_lib.translate_query(MagicMock(), MagicMock())

    def test_register_schema(self):
        # remove any existing registered schemas
        with patch('ckanext.versioned_datastore.lib.query.schema.schemas', {}):
            schema_lib.register_schema('v1.0.1', MagicMock())
            assert list(schema_lib.schemas.keys()) == ['v1.0.1']
            schema_lib.register_schema('v1.0.0', MagicMock())
            assert list(schema_lib.schemas.keys()) == ['v1.0.0', 'v1.0.1']
            schema_lib.register_schema('v2.0.1', MagicMock())
            assert list(schema_lib.schemas.keys()) == ['v1.0.0', 'v1.0.1', 'v2.0.1']



from ckan.plugins import interfaces


class IVersionedDatastore(interfaces.Interface):
    """
    Allow modifying versioned datastore logic.
    """

    def datastore_modify_data_dict(self, context, data_dict):
        """
        Allows modification of the data dict before it is validated and used to create
        the search object. This function should be used to remove/add/alter parameters
        in the data dict.

        :param context: the context
        :type context: dictionary
        :param data_dict: the parameters received from the user
        :type data_dict: dictionary
        """
        return data_dict

    def datastore_modify_search(self, context, original_data_dict, data_dict, search):
        """
        Allows modifications to the search before it is made. This is kind of analogous
        to IDatastore.datastore_search however instead of passing around a query dict,
        instead an elasticsearch-dsl Search object is passed.

        Each extension which implements this interface will be called in the order CKAN
        loaded them in, The search parameter will be the output of the previous
        extension's interface implementation, thus creating a chain of extensions, each
        getting a go at altering the search object if necessary. The base
        datastore_search function provides the initial search object.

        Implementors of this function should return the search object. Don't forget that
        most functions on the search object are chainable and create a copy of the
        search object - ensure you're returning the modified object!

        Two data dicts are passed into this function, the original data dict as it was
        before any ``datastore_modify_data_dict`` functions got to it, and the modified
        data_dict that was used to create the search object by the core functionality.
        This allows someone to, for example, remove a part of the data_dict in
        ``datastore_modify_data_dict`` to avoid it being added into the search object by
        the core functionality. Then, by implementing this function, they can add their
        custom search parts based on the details they removed by extracting them from
        the original_data_dict.

        :param context: the context
        :type context: dictionary
        :param original_data_dict: the parameters received from the user
        :type original_data_dict: dictionary
        :param data_dict: the parameters received from the user after they have been
                          modified by implementors of ``datastore_modify_data_dict``
        :type data_dict: dictionary
        :param search: the current search, as changed by the previous
                       IVersionedDatastore extensions in the chain
        :type search: elasticsearch-dsl Search object

        :returns: the search object with your modifications
        :rtype: elasticsearch-dsl Search object
        """
        return search

    def datastore_modify_result(self, context, original_data_dict, data_dict, result):
        """
        Allows modifications to the result after the search.

        Each extension which implements this interface will be called in the order CKAN
        loaded them in, The result parameter will be the output of the previous
        extension's interface implementation, thus creating a chain of extensions, each
        getting a go at altering the result object if necessary.

        Implementors of this function should return the result object so that the
        datastore_search function can build the final return dict.

        :param context: the context
        :type context: dictionary
        :param original_data_dict: the parameters received from the user
        :type original_data_dict: dictionary
        :param data_dict: the parameters received from the user after they have been
                          modified by implementors of ``datastore_modify_data_dict``
        :type data_dict: dictionary
        :param result: the current result, as changed by the previous
                       IVersionedDatastore extensions in the chain
        :type result: elasticsearch result object

        :returns: the result object with your modifications
        :rtype: elasticsearch result object
        """
        return result

    def datastore_modify_fields(self, resource_id, mapping, fields):
        """
        Allows modification of the field definitions before they are returned with the
        results of a datastore_search. The definitions are used in CKAN by the recline
        view and therefore need to abide by any of its requirements. By default all
        fields are included and are simply made up of a dict containing an id and type
        key. The id is the name of the field and the type is always string.

        :param resource_id: the resource id that was searched
        :param mapping: the mapping for the elasticsearch index containing the
                        resource's data. This is the raw mapping as a dict, retrieved
                        straight from elasticsearch's mapping endpoint
        :param fields: the field definitions that have so far been extracted from the
                       mapping, by default this is all fields
        :return: the list of field definition dicts
        """
        return fields

    def datastore_modify_index_doc(self, resource_id, index_doc):
        '''
        Action allowing the modification of a resource's data during indexing. The
        index_doc passed is a dict in the form:

            {
                "data": {},
                "meta": {}
            }

        which will be sent in this form to elasticsearch for indexing. The data key's
        value contains the data for the record at a version. The meta key's value
        contains metadata for the record so that we can search it correctly. Breakdown
        of the standard keys in the meta dict:

            - versions: a dict containing the range of versions this document is valid
                        for. This is represented using an elasticsearch range, with
                        "gte" for the  first valid version and "lt" for the last
                        version. If the "lt" key is missing the data is current.
            - version: the version of this record this data represents
            - next_version: will be missing if this data is current but if present, this
                            holds the value of the next version of this record

        If needed, the record id will be located in the index_doc under the key '_id'.

        :param resource_id: the id of the resource being indexed
        :param index_doc: a dict that will be sent to elasticsearch
        :return: the dict for elasticsearch to index
        '''
        return index_doc

    def datastore_is_read_only_resource(self, resource_id):
        """
        Allows implementors to designate certain resources as read only. This is purely
        a datastore side concept and should be used to prevent actions such as:

            - creating a new datastore for the resource (i.e. creating the index in
              elasticsearch)
            - upserting data into the datastore for this resource
            - deleting the datastore for this resource
            - reindexing data in the datastore for this resource

        :param resource_id: the resource id to check
        :return: True if the resource should be treated as read only, False if not
        """
        return False

    def datastore_after_indexing(self, request, splitgill_stats, stats_id):
        """
        Allows implementors to hook onto the completion of an indexing task. This
        function doesn't return anything and any exceptions it raises will be caught and
        ignored.

        :param request: the ResourceIndexRequest object that triggered the indexing task
        :param splitgill_stats: the statistics about the indexing task from splitgill
        :param stats_id: the id of the statistics entry in the ImportStats database
                         table
        """
        pass

    def datastore_reserve_slugs(self):
        """
        Allows implementors to reserve queries using reserved pretty slugs. Implementors
        should return a dict made up of reserved pretty slugs as keys and then the slug
        parameters as the values. These values should be another dict containing the
        following optional keys:

            - query, a dict query (defaults to {})
            - query_version, the query schema version (defaults to the latest query
                             schema version)
            - version, the version of the data to search at (defaults to None)
            - resource_ids, a list of resource ids to search (defaults to all resource
                            ids)
            - resource_ids_and_versions, a dict of resource ids and specific versions to
                                         search at (defaults to an empty dict)

        If a slug already exists in the database with the same reserved pretty slug and
        the same query parameters then nothing happens.

        If a slug already exists in the database with the same reserved pretty slug but
        a different set of query parameters then a DuplicateSlugException is raised.

        If a slug already exists in the database with the same query parameters but no
        reserved pretty slug then the reserved pretty slug is added to the slug.
        """
        return {}

    def datastore_modify_guess_fields(self, resource_ids, fields):
        """
        Allows plugins to manipulate the Fields object used to figure out the groups
        that should be returned by the datastore_guess_fields action.

        :param resource_ids: a list of resource ids
        :param fields: a Fields object
        :return: the Fields object
        """
        return fields

    def datastore_multisearch_modify_response(self, response):
        """
        Allows plugins to alter the response dict returned from the
        datastore_multisearch action before it is returned.

        :param response: the dict to be returned to the caller
        :return: the response dict
        """
        return response

    def datastore_before_convert_basic_query(self, basic_query):
        """
        Allows plugins to modify a basic query (probably taken from a URL), e.g. to
        remove custom filters before processing.

        :param basic_query: the query dict to be modified
        :return: the modified query
        """
        return basic_query

    def datastore_after_convert_basic_query(self, basic_query, multisearch_query):
        """
        Allows plugins to modify a converted query, e.g. to add back in any complex
        custom filters.

        :param basic_query: the original basic query, before it was modified by other
                            plugins
        :param multisearch_query: the converted multisearch version of the query
        :return: the modified multisearch query
        """
        return multisearch_query


class IVersionedDatastoreQuerySchema(interfaces.Interface):
    def get_query_schemas(self):
        """
        Hook to allow registering custom query schemas.

        :return: a list of tuples of the format (query schema version, schema object)
                 where the query schema version is a string of format v#.#.# and the
                 schema object is an instance of
                 ckanext.versioned_datastore.lib.query.Schema
        """
        return []


class IVersionedDatastoreDownloads(interfaces.Interface):
    def download_modify_notifier_start_templates(
        self, text_template: str, html_template: str
    ):
        """
        Hook allowing other extensions to modify the templates used when sending
        notifications that a download has started. The templates can be modified in
        place or completely replaced.

        :param text_template: the text template string
        :param html_template: the html template string
        :return: a 2-tuple containing the text template string and the html template
                 string
        """
        return text_template, html_template

    def download_modify_notifier_end_templates(
        self, text_template: str, html_template: str
    ):
        """
        Hook allowing other extensions to modify the templates used when sending
        notifications that a download has completed successfully. The templates can be
        modified in place or completely replaced.

        :param text_template: the text template string
        :param html_template: the html template string
        :return: a 2-tuple containing the text template string and the html template
                 string
        """
        return text_template, html_template

    def download_modify_notifier_error_templates(
        self, text_template: str, html_template: str
    ):
        """
        Hook allowing other extensions to modify the templates used when sending
        notifications that a download has failed. The templates can be modified in place
        or completely replaced.

        :param text_template: the text template string
        :param html_template: the html template string
        :return: a 2-tuple containing the text template string and the html template
                 string
        """
        return text_template, html_template

    def download_modify_notifier_template_context(self, request, context):
        """
        Hook allowing other extensions to modify the templating context used to generate
        the download email (both plain text and HTML versions) before it is sent.

        The default context contains:
            - "download_url": the download zip's full URL
            - "site_url": the CKAN site's full URL (this is taken straight from the
                          config)

        :param request: the DownloadRequest object
        :param context: templating context dict
        :return: context templating dict
        """
        return context

    def download_derivative_generators(self, registered_derivatives=None):
        """
        Extend or modify the list of derivative generators.

        :param registered_derivatives: a dict of existing derivative generator classes,
                                       returned from previously loaded plugins
        :return: a dict of loaded derivative generator classes, keyed on the name used
                 to specify them in download requests
        """
        return registered_derivatives or {}

    def download_file_servers(self, registered_servers=None):
        """
        Extend or modify the list of file servers.

        :param registered_servers: a dict of existing file server classes, returned from
                                   previously loaded plugins
        :return: a dict of loaded file server classes, keyed on the name used to specify
                 them in download requests
        """
        return registered_servers or {}

    def download_notifiers(self, registered_notifiers=None):
        """
        Extend or modify the list of download notifiers.

        :param registered_notifiers: a dict of existing notifier classes, returned from
                                     previously loaded plugins
        :return: a dict of loaded notifier classes, keyed on the name used to specify
                 them in download requests
        """
        return registered_notifiers or {}

    def download_data_transformations(self, registered_transformations=None):
        """
        Extend or modify the list of data transformations.

        :param registered_transformations: a dict of existing data transformations,
                                           returned from previously loaded plugins
        :return: a dict of loaded transformations, keyed on the name used to specify
                 them in download requests
        """
        return registered_transformations or {}

    def download_before_run(
        self, query_args, derivative_args, server_args, notifier_args
    ):
        """
        Hook allowing other extensions to modify args before any search is run or files
        generated.
        FIXME: this should be renamed to download_before_init or similar

        :param query_args: a QueryArgs object
        :param derivative_args: a DerivativeArgs object
        :param server_args: a ServerArgs object
        :param notifier_args: a NotifierArgs object
        :return: all four args objects (query_args, derivative_args, server_args,
                 notifier_args)
        """
        return query_args, derivative_args, server_args, notifier_args

    def download_after_init(self, query):
        """
        Hook notifying that the downloader and associated objects (e.g. the query) have
        been initialised. Does not allow modification; purely for notification purposes.

        :param query: the query for this download
        :return: None
        """
        return

    def download_modify_manifest(self, manifest, request):
        """
        Hook allowing other extensions to modify the manifest before the download file
        is written. Modifications to the request object are not saved.

        :param manifest: the manifest dict
        :param request: the DownloadRequest object
        :return: the manifest dict
        """
        return manifest

    def download_modify_eml(self, eml_dict, query):
        """
        Hook allowing other extensions to modify the content of the EML before it's
        transformed into XML and written to file.

        :param eml_dict: the current eml content, as a dict
        :param query: the query for this download
        :return: the modified eml dict
        """
        return eml_dict

    def download_after_run(self, request):
        """
        Hook notifying that a download has finished (whether failed or completed). Does
        not allow modification; purely for notification purposes.

        :param request: the DownloadRequest object
        :return: None
        """
        return



from splitgill.ingestion.feeders import BaseRecord


def _convert(data):
    """
    Converts the given dict into a suitable format for storage in mongo. For us this
    means:

        - replacing '.' with '_' as neither mongo nor elasticsearch can handle dots in field
          names as they both use the dot notation for nested field access
        - ignoring fields with no name (we use a falsey check on the field name) to ensure we
          don't create fields that are the empty string

    This function is recursive.

    :return: a dict ready for storage in mongo
    """
    converted = {}
    for field, value in data.items():
        # elasticsearch doesn't allow empty fields, plus it's silly so ignore them
        if not field:
            continue
        # mongo doesn't allow dots in keys so replace them with underscores
        field = field.replace('.', '_')
        if isinstance(value, dict):
            converted[field] = _convert(value)
        else:
            converted[field] = value
    return converted


class DatastoreRecord(BaseRecord):
    """
    Represents a record from a feeder which needs to be ingested into mongo.
    """

    def __init__(self, version, record_id, data, resource_id):
        '''
        :param version: the version of this record
        :param record_id: the record's id
        :param data: a dict containing the fields and values for the record
        :param resource_id: the resource id this record belongs to
        '''
        super(DatastoreRecord, self).__init__(version)
        self.record_id = record_id
        self.data = data
        self.resource_id = resource_id

    def convert(self):
        """
        Converts the record into a suitable format for storage in mongo. For us this
        means:

            - replacing '.' with '_' as neither mongo nor elasticsearch can handle dots in field
              names as they both use the dot notation for nested field access
            - ignoring fields with no name (we use a falsey check on the field name) to ensure we
              don't create fields that are the empty string

        :return: a dict ready for storage in mongo
        """
        return _convert(self.data)

    @property
    def id(self):
        """
        Returns the id of the record.

        :return: the id
        """
        return self.record_id

    @property
    def mongo_collection(self):
        """
        Returns the name of the collection in mongo which should store this record's
        data.

        :return: the name of the target mongo collection
        """
        return self.resource_id



import json
import logging
from contextlib import suppress
from logging.handlers import RotatingFileHandler

from ckan.plugins import toolkit

config_base = 'ckanext.versioned_datastore.query_log_{}'
# default enabled to False
is_enabled = toolkit.asbool(toolkit.config.get(config_base.format('enabled'), False))
# default path to None which, if the enabled flag is True, will direct everything to the console log
path = toolkit.config.get(config_base.format('path'), None)
# default the rotate at bytes value to 5MiB
rotate_at_bytes = toolkit.asint(
    toolkit.config.get(config_base.format('rotate_at_bytes'), 5242880)
)
# default the number of backup files to keep at 5
backup_count = toolkit.asint(toolkit.config.get(config_base.format('backup_count'), 5))


if is_enabled:
    logger = logging.getLogger('query-log')
    logger.setLevel(logging.INFO)
    if path:
        file_handler = RotatingFileHandler(
            path, encoding='utf-8', maxBytes=rotate_at_bytes, backupCount=backup_count
        )
        formatter = logging.Formatter('%(asctime)s: %(message)s')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        # this stops the query log messages from coming out in the main console log
        logger.propagate = False
else:
    logger = None


def log_query(query, source):
    """
    Call this to log a query dict to the query log file (if the config permits it). The
    query is simply json dumped. If query logging is disabled then nothing happens.

    :param query: the query dict
    :param source: the source of the query (e.g. multisearch or basicsearch)
    """
    if is_enabled:
        # use suppress just to make sure nothing explodes whilst logging
        with suppress(Exception):
            logger.info(
                f'{source}: {json.dumps(query, sort_keys=True, ensure_ascii=False)}'
            )



import pytest
from mock import patch

from ckanext.versioned_datastore.logic.actions.downloads import datastore_queue_download
from ckanext.versioned_datastore.logic.actions.meta.arg_objects import (
    QueryArgs,
    DerivativeArgs,
    NotifierArgs,
)
from tests.helpers import patches


class TestQueueDownload:
    @pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
    @pytest.mark.usefixtures('with_plugins', 'with_versioned_datastore_tables')
    def test_queue_direct_call(self):
        # there is a very similar test in test_downloads.py that calls this via the API
        # instead
        resource_ids = ['test-resource-id']

        with patch(
            'ckan.plugins.toolkit.enqueue_job'
        ) as enqueue_mock, patches.rounded_versions(), patches.get_available_resources(
            resource_ids
        ), patches.url_for():
            datastore_queue_download(
                {},
                QueryArgs(),
                DerivativeArgs(format='csv'),
                notifier=NotifierArgs(type='none'),
            )
            assert enqueue_mock.call_count == 1



"""
Add server_args to download request model.

Revision ID: d2ca5da0573f
Revises: 526b12c69d55
Create Date: 2023-06-28 13:28:33.607360
"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB


# revision identifiers, used by Alembic.
revision = 'd2ca5da0573f'
down_revision = '526b12c69d55'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'vds_download_request', sa.Column('server_args', JSONB, nullable=True)
    )


def downgrade():
    op.drop_column('vds_download_request', 'server_args')



import logging

from ckan.plugins import toolkit

from .base import BaseTransform

log = logging.getLogger(__name__)
base_url = toolkit.config.get('ckan.site_url')
object_endpoint = toolkit.config.get(
    'ckanext.versioned_datastore.record_view_endpoint', 'object.view'
)


class IdAsUrlTransform(BaseTransform):
    name = 'id_as_url'

    def __init__(self, field=None, **kwargs):
        """
        :param field: the name of the data field that contains the ID and that will contain the URL
        """
        super().__init__(**kwargs)
        self.field = field

    def __call__(self, data):
        """
        Reformat an ID field as a URL (probably one that links to that record). Requires
        an endpoint (config option ckanext.versioned_datastore.record_view_endpoint,
        default 'object.view') taking the ID field as the 'uuid' named argument.

        :param data: the record data to be transformed
        :return: the transformed data (or untransformed if there was an error).
        """
        if self.field is None:
            return data
        try:
            object_id = data.get(self.field)
            if object_id is None or object_id == '':
                log.error(f'Failed to get uuid from field "{self.field}".')
                return data
            kwargs = {'uuid': object_id}
            url = toolkit.url_for(object_endpoint, **kwargs)
        except:
            log.error(f'Failed to generate URL from ID.', exc_info=True)
            return data
        data[self.field] = base_url + url
        return data



import copy
import json
from elasticsearch import NotFoundError
from elasticsearch_dsl import MultiSearch, Search
from splitgill.indexing.utils import DOC_TYPE
from splitgill.search import create_version_query

from ckan.lib.search import SearchIndexError
from ckan.plugins import PluginImplementations
from .. import common
from ..datastore_utils import prefix_resource, prefix_field
from ..importing.details import get_all_details
from ...interfaces import IVersionedDatastore


def run_search(search, indexes, version=None):
    """
    Convenience function to runs a search on the given indexes using the client
    available in this module.

    If the index(es) required for the search are missing then a CKAN SearchIndexError exception is
    raised.

    :param search: the elasticsearch-dsl search object
    :param indexes: either a list of index names to search in or a single index name as a string
    :param version: version to filter the search results to, optional
    :return: the result of running the query
    """
    try:
        if version is not None:
            search = search.filter(create_version_query(version))
        if isinstance(indexes, str):
            indexes = [indexes]
        return search.index(indexes).using(common.ES_CLIENT).execute()
    except NotFoundError as e:
        raise SearchIndexError(e.error)


def format_facets(aggs):
    '''
    Formats the facet aggregation result into the format we require. Specifically we expand the
    buckets out into a dict that looks like this:

        {
            "facet1": {
                "details": {
                    "sum_other_doc_count": 34,
                    "doc_count_error_upper_bound": 3
                },
                "values": {
                    "value1": 1,
                    "value2": 4,
                    "value3": 1,
                    "value4": 2,
                }
            },
            etc
        }

    etc.

    :param aggs: the aggregation dict returned from splitgill/elasticsearch
    :return: the facet information as a dict
    '''
    facets = {}
    for facet, details in aggs.items():
        facets[facet] = {
            'details': {
                'sum_other_doc_count': details['sum_other_doc_count'],
                'doc_count_error_upper_bound': details['doc_count_error_upper_bound'],
            },
            'values': {
                value_details['key']: value_details['doc_count']
                for value_details in details['buckets']
            },
        }

    return facets


# this dict stores cached get_field returns. It is only cleared by restarting the server. This is
# safe because the cached data is keyed on the rounded version and is therefore stable as old
# versions of data can't be modified, so the fields will always be valid. If for some reason this
# isn't the case (such as if redactions for specific fields get added later and old versions of
# records are updated) then the server just needs a restart and that's it).
field_cache = {}


def get_fields(resource_id, version=None):
    """
    Given a resource id, returns the fields that existed at the given version. If the
    version is None then the fields for the latest version are returned.

    The response format is important as it must match the requirements of reclineJS's field
    definitions. See http://okfnlabs.org/recline/docs/models.html#field for more details.

    All fields are returned by default as string types. This is because we have the capability to
    allow searchers to specify whether to treat a field as a string or a number when searching and
    therefore we don't need to try and guess the type and we can leave it to the user to know the
    type which won't cause problems like interpreting a field as a number when it shouldn't be (for
    example a barcode like '013655395'). If we decide that we do want to work out the type we simply
    need to add another step to this function where we count how many records in the version have
    the '.number' subfield - if the number is the same as the normal field count then the field is a
    number type, if not it's a string.

    The fields are returned in either alphabetical order, or if we have the ingestion details for
    the resource at the required version then the order of the fields will match the order of the
    fields in the original source.

    :param resource_id: the resource's id
    :param version: the version of the data we're querying (default: None, which means latest)
    :return: a list of dicts containing the field data
    """
    # figure out the index name from the resource id
    index = prefix_resource(resource_id)
    # figure out the rounded version so that we can figure out the fields at the right version
    rounded_version = common.SEARCH_HELPER.get_rounded_versions([index], version)[index]
    # the key for caching should be unique to the resource and the rounded version
    cache_key = (resource_id, rounded_version)

    # if there is a cached version, return it! Woo!
    if cache_key in field_cache:
        return field_cache[cache_key]

    # create a list of field details, starting with the always present _id field
    fields = [{'id': '_id', 'type': 'integer'}]
    # lookup the mapping on elasticsearch to get all the field names
    mapping = common.ES_CLIENT.indices.get_mapping(index)[index]
    # if the rounded version response is None that means there are no versions available which
    # shouldn't happen, but in case it does for some reason, just return the fields we have
    # already
    if rounded_version is None:
        return mapping, fields

    # retrieve all the resource's details up to the target version to get the column orders at each
    # version as they were in the ingestion sources for each version
    all_details = get_all_details(resource_id, up_to_version=version)
    # this set is used to avoid duplicating fields, we preload it with the _id column because we
    # want to ignore that (it's already in the fields list defined above)
    seen_fields = {'_id'}
    field_names = []

    if all_details:
        # the all_details variable is an OrderedDict in ascending version order. We want to iterate
        # in descending version order though so that we respect the column order at the version
        # we're at before respecting any data from previous versions
        for details in reversed(all_details.values()):
            columns = [
                column for column in details.get_columns() if column not in seen_fields
            ]
            field_names.extend(columns)
            seen_fields.update(columns)

    mapped_fields = mapping['mappings'][DOC_TYPE]['properties']['data']['properties']
    # add any unseen mapped fields to the list of names. If we have a details object for each
    # version this shouldn't add any additional fields and if not it ensures we don't miss any
    field_names.extend(
        field for field in sorted(mapped_fields) if field not in seen_fields
    )

    if field_names:
        # find out which fields exist in this version and how many values each has
        search = MultiSearch(using=common.ES_CLIENT, index=index)
        for field in field_names:
            # create a search which finds the documents that have a value for the given field at the
            # rounded version. We're only interested in the counts though so set size to 0
            search = search.add(
                Search()
                .extra(size=0)
                .filter('exists', **{'field': prefix_field(field)})
                .filter('term', **{'meta.versions': rounded_version})
            )

        # run the search and get the response
        responses = search.execute()
        for i, response in enumerate(responses):
            # if the field has documents then it should be included in the fields list
            if response.hits.total > 0:
                fields.append(
                    {
                        'id': field_names[i],
                        # by default everything is a string
                        'type': 'string',
                    }
                )

    # stick the result in the cache for next time
    field_cache[cache_key] = (mapping, fields)

    return mapping, fields


def convert_to_multisearch(query):
    # save a copy of the original query
    basic_query = copy.deepcopy(query)
    multisearch_query = {}

    # allow other plugins to modify the query before processing, e.g. to remove any
    # custom filters
    for plugin in PluginImplementations(IVersionedDatastore):
        query = plugin.datastore_before_convert_basic_query(query)

    if 'q' in query:
        multisearch_query['search'] = query['q']

    if 'filters' in query:
        filter_list = []
        for field, values in query['filters'].items():
            if not isinstance(values, list):
                values = [values]
            if field == '__geo__':
                for value in values:
                    if isinstance(value, str):
                        value = json.loads(value)
                    if value['type'] == 'Polygon':
                        filter_list.append({'geo_custom_area': [value['coordinates']]})
                    else:
                        # I cannot find any examples of anything other than polygons, so
                        # I'm not sure it was ever implemented for the old searches
                        raise NotImplemented
            else:
                subgroup = []
                subgroup_count = 0
                for value in values:
                    if field != '' and value != '':
                        subgroup.append(
                            {'string_equals': {'fields': [field], 'value': value}}
                        )
                        subgroup_count += 1
                if subgroup_count > 1:
                    filter_list.append({'or': subgroup})
                elif subgroup_count == 1:
                    filter_list += subgroup
        multisearch_query['filters'] = {'and': filter_list}

    # allow plugins to modify the output, with the additional context of the original
    # basic query
    for plugin in PluginImplementations(IVersionedDatastore):
        multisearch_query = plugin.datastore_after_convert_basic_query(
            basic_query, multisearch_query
        )

    return multisearch_query
